/***	morph_main.cc	***/

/*	Copyright (C) Jan Daciuk, 1996	*/

/*

This program looks for words from standard input, and tries to find
morphological information for that words in speficied dictionaries.

Synopsis:
tr_morph [-d dictionary]... [-d dictionary]...

where dictionary is a file containing the dictionary in a form of a binary
transducer. At least one dictionay must be specified.

Words are read from standard input, and results written to standard
output.

*/

#include	<iostream>
#include	<fstream>
#include	<string.h>
#include	<stdlib.h>
#include	<new>
#include	"tr.h"
#include	"nstr.h"
#include	"commont.h"
#include	"morpht.h"
#include	"tr_version.h"

static const int	MAX_LINE_LENGTH = 512; // input data max line length


int
main(const int argc, const char *argv[]);
int
usage(const char *prog_name);
void
not_enough_memory(void);



/* Name:	not_enough_memory
 * Class:	None.
 * Purpose:	Inform the user that there is not enough memory to continue
 *		and finish the program.
 * Parameters:	None.
 * Returns:	Nothing.
 * Remarks:	None.
 */
void
not_enough_memory(void)
{
  cerr << "Not enough memory for the transducer\n";
  exit(4);
}//not_enough_memory

/* Name:	main
 * Class:	None.
 * Purpose:	Launches the program.
 * Parameters:	argc		- (i) number of program arguments;
 *		argv		- (i) program arguments;
 * Returns:	Program exit code:
 *		0	- OK;
 *		1	- invalid options;
 *		2	- dictionary file could not be opened;
 * Remarks:	None.
 */
int
main(const int argc, const char *argv[])
{
  word_list	dict;		// dictionary file name 
  int		arg_index;	// current argument number
  word_list	inputs;		// names of input files (if any)
  const char	*lang_file = NULL;	// language file name
  int		generate = FALSE; // whether to generate instead of analysing
  char		annot_sep = '\t'; // separates annotations (tags) from lexemes

  set_new_handler(&not_enough_memory);


  for (arg_index = 1; arg_index < argc; arg_index++) {
    if (argv[arg_index][0] != '-')
      // not an option
      return usage(argv[0]);
    if (argv[arg_index][1] == 'd') {
      // dictionary file name
      if (++arg_index >= argc)
	return usage(argv[0]);
      dict.insert(argv[arg_index]);
    }
    else if (argv[arg_index][1] == 'i') {
      // input file name
      if (++arg_index >= argc)
	return usage(argv[0]);
      inputs.insert(argv[arg_index]);
    }
    else if (argv[arg_index][1] == 'l') {
      // language file name
      if (++arg_index >= argc)
	return usage(argv[0]);
      lang_file = argv[arg_index];
    }
    else if (argv[arg_index][1] == 'G') {
      generate = TRUE;
    }
    else if (argv[arg_index][1] == 'A') {
      if (++arg_index >= argc)
	return usage(argv[0]);
      annot_sep = argv[arg_index][0];
    }
    else if (argv[arg_index][1] == 'v') {
      // display version details
      cout << VERSION << endl;
#include "comp_options.h"
      return 0;
    }
    else {
      cerr << argv[0] << ": unrecognized option\n";
      return usage(argv[0]);
    }
  }//for

  if (dict.how_many() == 0) {
    cerr << argv[0] << ": at least one dictionary must be specified\n";
    return usage(argv[0]);
  }

  dict.reset();
  morph_tr fsa_dict(&dict, lang_file);
  if (!fsa_dict) {
    if (inputs.how_many()) {
      inputs.reset();
      do {
	ifstream iff(inputs.item());
	if (iff) {
	  tr_io io_obj(&iff, cout, MAX_LINE_LENGTH);
	  if (generate)
	    fsa_dict.gen_words_from_file(io_obj, annot_sep);
	  else
	    fsa_dict.morph_file(io_obj);
	}
	else {
	  cerr << argv[0] << ": Could not open input file " << inputs.item() << endl;
	}
      } while (inputs.next());
      return 0;
    }
    else {
      tr_io io_obj(&cin, cout, MAX_LINE_LENGTH);
      if (generate)
	fsa_dict.gen_words_from_file(io_obj, annot_sep);
      else
	return fsa_dict.morph_file(io_obj);
    }
  }
  else
    return fsa_dict;
}//main

/* Name:	usage
 * Class:	None.
 * Purpose:	Prints program synopsis.
 * Parameters:	prog_name	- (i) program name.
 * Returns:	1.
 * Remarks:	None.
 */
int
usage(const char *prog_name)
{
  cerr << "Usage:\n" << prog_name << " [options]... [<infile] [>outfile]\n"
       << "Options:\n"
       << "-d dictionary\t- transducer file (multiple files allowed)\n"
       << "-i input_file_name (multiple files allowed)\n"
       << "-l language_file\t- file that defines characters allowed in words\n"
       << "\tand case conversions\n"
       << "\t[default: ASCII letters, standard conversions]\n"
       << "-G\t- generate inflected forms (from lexemes and tags) "
       << "instead of analysing them\n"
       << "-A annot_sep\t- separates lexemes from tags "
       << "in input with -G option\n"
       << "\t[default: HT (horizontal tabulation]\n"
       << "-v\t	- version details\n"
       << "Standard output is used for displaying results.\n"
       << "At least one dictionary must be present.\n";
  return 1;
}//usage

/***	EOF morph_main.cc	***/
